"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.useCollection = void 0;

var _react = require("react");

var _reactQuery = require("react-query");

var _Provider = require("./Provider");

var _Cache = require("./Cache");

var _types = require("./types");

var _utils = require("./utils");

const createFirestoreRef = (firestore, path, {
  where,
  orderBy,
  limit,
  startAt,
  endAt,
  startAfter,
  endBefore,
  isCollectionGroup
}) => {
  let ref = firestore.collection(path);

  if (isCollectionGroup) {
    ref = firestore.collectionGroup(path);
  }

  if (where) {
    function multipleConditions(w) {
      return !!w && Array.isArray(w[0]);
    }

    if (multipleConditions(where)) {
      where.forEach(w => {
        ref = ref.where(w[0], w[1], w[2]);
      });
    } else if (typeof where[0] === "string" && typeof where[1] === "string") {
      ref = ref.where(where[0], where[1], where[2]);
    }
  }

  if (orderBy) {
    if (typeof orderBy === "string") {
      ref = ref.orderBy(orderBy);
    } else if (Array.isArray(orderBy)) {
      function multipleOrderBy(o) {
        return Array.isArray(o[0]);
      }

      if (multipleOrderBy(orderBy)) {
        orderBy.forEach(([order, direction]) => {
          ref = ref.orderBy(order, direction);
        });
      } else {
        const [order, direction] = orderBy;
        ref = ref.orderBy(order, direction);
      }
    }
  }

  if (startAt) {
    ref = ref.startAt(startAt);
  }

  if (endAt) {
    ref = ref.endAt(endAt);
  }

  if (startAfter) {
    ref = ref.startAfter(startAfter);
  }

  if (endBefore) {
    ref = ref.endBefore(endBefore);
  }

  if (limit) {
    ref = ref.limit(limit);
  }

  return ref;
};

const createListenerAsync = async (firestore, queryClient, path = undefined, queryString, ignoreFirestoreDocumentSnapshotField, setHasNextPage) => {
  return new Promise((resolve, reject) => {
    var _JSON$parse;

    if (!path) {
      return resolve({
        initialData: [],
        unsubscribe: _types.empty.function
      });
    }

    const query = (_JSON$parse = JSON.parse(queryString)) !== null && _JSON$parse !== void 0 ? _JSON$parse : {};
    const ref = createFirestoreRef(firestore, path, query);
    const unsubscribe = ref.onSnapshot({
      includeMetadataChanges: true
    }, {
      next: querySnapshot => {
        const data = [];
        querySnapshot.forEach(doc => {
          var _doc$data;

          const docData = (_doc$data = doc.data()) !== null && _doc$data !== void 0 ? _doc$data : _types.empty.object;
          (0, _utils.parseDates)(docData);
          const docToAdd = { ...docData,
            id: doc.id,
            exists: doc.exists,
            hasPendingWrites: doc.metadata.hasPendingWrites,
            __snapshot: ignoreFirestoreDocumentSnapshotField ? undefined : doc
          }; // update individual docs in the cache

          queryClient.setQueryData(doc.ref.path, docToAdd);
          data.push(docToAdd);
        });
        setHasNextPage(!querySnapshot.empty); // resolve initial data

        resolve({
          initialData: data,
          unsubscribe
        }); // update on listener fire

        queryClient.setQueryData([path, queryString], data);
      },
      error: reject
    });
  });
};
/**
 * Call a Firestore Collection
 * @template Doc
 * @param path String if the document is ready. If it's not ready yet, pass `null`, and the request won't start yet.
 * @param [query] - Dictionary with options to query the collection.
 * @param [options] - takes any of useQuery options.
 */


const useCollection = (path, options, query) => {
  const [isFetchingNextPage, setIsFetchingNextPage] = (0, _react.useState)(false);
  const [hasNextPage, setHasNextPage] = (0, _react.useState)(true);
  const {
    firestore
  } = (0, _Provider.useFirestore)();
  const queryClient = (0, _reactQuery.useQueryClient)();
  const unsubscribeRef = (0, _react.useRef)(null);
  const {
    where,
    endAt,
    endBefore,
    startAfter,
    startAt,
    orderBy,
    limit,
    isCollectionGroup,
    ignoreFirestoreDocumentSnapshotField = true
  } = query || {}; // why not just put this into the ref directly?
  // so that we can use the useEffect down below that triggers revalidate()

  const memoQueryString = (0, _react.useMemo)(() => JSON.stringify({
    where,
    endAt,
    endBefore,
    startAfter,
    startAt,
    orderBy,
    limit,
    isCollectionGroup
  }), [endAt, endBefore, isCollectionGroup, limit, orderBy, startAfter, startAt, where]);

  async function fetch() {
    if (unsubscribeRef.current) {
      unsubscribeRef.current();
      unsubscribeRef.current = null;
    }

    const {
      unsubscribe,
      initialData
    } = await createListenerAsync(firestore, queryClient, path, memoQueryString, ignoreFirestoreDocumentSnapshotField, setHasNextPage);
    unsubscribeRef.current = unsubscribe;
    return initialData;
  }

  const {
    data,
    status,
    error
  } = (0, _reactQuery.useQuery)([path, memoQueryString], fetch, { ...options,
    notifyOnChangeProps: "tracked"
  });
  const {
    mutateAsync
  } = (0, _reactQuery.useMutation)(async ({
    data: newData,
    subPath
  }) => {
    if (!path) return Promise.resolve([]);
    const newPath = subPath ? path + "/" + subPath : path;
    const dataArray = Array.isArray(newData) ? newData : [newData];
    const ref = firestore.collection(newPath);
    const docsToAdd = dataArray.map(doc => ({ ...doc,
      // generate IDs we can use that in the local cache that match the server
      id: ref.doc().id
    })); // add to network

    const batch = firestore.batch();
    docsToAdd.forEach(({
      id,
      ...doc
    }) => {
      // take the ID out of the document
      batch.set(ref.doc(id), doc);
    });
    await batch.commit();
    return Promise.resolve(docsToAdd);
  }, {
    // Always refetch after error or success:
    onSettled: () => {
      queryClient.invalidateQueries([path, memoQueryString]);
    }
  });
  (0, _react.useEffect)(() => {
    //should it go before the useQuery?
    return () => {
      // clean up listener on unmount if it exists
      if (unsubscribeRef.current) {
        unsubscribeRef.current();
        unsubscribeRef.current = null;
      }
    }; // should depend on the path, queyr being the same...
  }, [path, memoQueryString]); // add the collection to the cache,
  // so that we can mutate it from document calls later

  (0, _react.useEffect)(() => {
    if (path) _Cache.collectionCache.addCollectionToCache(path, memoQueryString);
  }, [path, memoQueryString]);
  /**
   * `add(data, subPath?)`: Extends the Firestore document [`add` function](https://firebase.google.com/docs/firestore/manage-data/add-data).
   * - It also updates the local cache using react-query's `setQueryData`. This will prove highly convenient over the regular `add` function provided by Firestore.
   * - If the second argument is defined it will be concatinated to path arg as a prefix
   */

  const add = async (newData, subPath) => mutateAsync({
    data: newData,
    subPath
  });

  const setCache = (0, _react.useCallback)(cachedData => {
    queryClient.setQueryData([path, memoQueryString], prevState => {
      if (!prevState) return [];
      return [...prevState, ...cachedData];
    });
  }, // eslint-disable-next-line react-hooks/exhaustive-deps
  [path, memoQueryString]);

  const fetchNextPage = async () => {
    if (!path || !(data !== null && data !== void 0 && data.length)) return;
    setIsFetchingNextPage(true); // get the snapshot of last document we have right now in our query

    const startAfterDocument = data[data.length - 1].__snapshot;
    const ref = createFirestoreRef(firestore, path, JSON.parse(memoQueryString)); // get more documents, after the most recent one we have

    const querySnapshot = await ref.startAfter(startAfterDocument).get();
    setHasNextPage(!querySnapshot.empty);
    const moreDocs = [];
    querySnapshot.docs.forEach(doc => {
      var _doc$data2;

      const docData = (_doc$data2 = doc.data()) !== null && _doc$data2 !== void 0 ? _doc$data2 : _types.empty.object;
      const docToAdd = { ...docData,
        id: doc.id,
        exists: doc.exists,
        hasPendingWrites: doc.metadata.hasPendingWrites,
        __snapshot: ignoreFirestoreDocumentSnapshotField ? undefined : doc
      }; // update individual docs in the cache

      queryClient.setQueryData(doc.ref.path, docToAdd);
      moreDocs.push(docToAdd);
    }); // mutate our local cache, adding the docs we just added

    setCache(moreDocs);
    setIsFetchingNextPage(false);
  };

  return {
    data,
    status,
    error,
    add,
    setCache,
    fetchNextPage,
    isFetchingNextPage,
    hasNextPage,

    /**
     * A function that, when called, unsubscribes the Firestore listener.
     *
     * The function can be null, so make sure to check that it exists before calling it.
     *
     * Note: This is not necessary to use. `useCollection` already unmounts the listener for you. This is only intended if you want to unsubscribe on your own.
     */
    unsubscribe: unsubscribeRef.current
  };
};

exports.useCollection = useCollection;
//# sourceMappingURL=useCollection.js.map