"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.useDocument = void 0;

var _react = require("react");

var _reactQuery = require("react-query");

var _Cache = require("./Cache");

var _types = require("./types");

var _useHelpers = require("./useHelpers");

var _Provider = require("./Provider");

var _utils = require("./utils");

function updateCollectionCache(queryClient, path, docId, data) {
  let collection = path.split(`/${docId}`).filter(Boolean);
  collection = collection.join("/");

  if (collection) {
    _Cache.collectionCache.getKeysFromCollectionPath(collection).forEach(key => {
      queryClient.setQueryData(key, (currentState = _types.empty.array) => {
        // don't mutate the current state if it doesn't include this doc
        if (!currentState.some(currDoc => currDoc.id && currDoc.id === data.id)) {
          return currentState;
        }

        return currentState.map(document => {
          if (document.id === data.id) {
            return data;
          }

          return document;
        });
      });
    });
  }
}

const createListenerAsync = async (firestore, queryClient, path) => {
  if (!path) {
    return {
      unsubscribe: _types.empty.function,
      initialData: _types.empty.object
    };
  }

  return await new Promise((resolve, reject) => {
    const unsubscribe = firestore.doc(path).onSnapshot({
      includeMetadataChanges: true
    }, doc => {
      var _doc$data;

      const docData = (_doc$data = doc.data()) !== null && _doc$data !== void 0 ? _doc$data : _types.empty.object;
      (0, _utils.parseDates)(docData);
      const data = { ...docData,
        id: doc.id,
        exists: doc.exists,
        hasPendingWrites: doc.metadata.hasPendingWrites
      };

      if (!data.hasPendingWrites) {
        queryClient.setQueryData(path, data);
        updateCollectionCache(queryClient, path, doc.id, data);
        resolve({
          initialData: data,
          unsubscribe
        });
      }
    }, reject);
  });
};

const useDocument = (path, options) => {
  const {
    firestore
  } = (0, _Provider.useFirestore)();
  const {
    deleteDocument: deleteDoc
  } = (0, _useHelpers.useHelpers)();
  const queryClient = (0, _reactQuery.useQueryClient)();
  const unsubscribeRef = (0, _react.useRef)(null);
  (0, _react.useEffect)(() => {
    return () => {
      // clean up listener on unmount if it exists
      if (unsubscribeRef.current) {
        unsubscribeRef.current();
        unsubscribeRef.current = null;
      }
    };
  }, [path]);

  async function fetch() {
    if (unsubscribeRef.current) {
      unsubscribeRef.current();
      unsubscribeRef.current = null;
    }

    const listner = await createListenerAsync(firestore, queryClient, path);
    unsubscribeRef.current = listner.unsubscribe;
    return listner.initialData;
  }

  const {
    data,
    status,
    error,
    refetch
  } = (0, _reactQuery.useQuery)(path || "", fetch, { ...options,
    enabled: !!path && (options === null || options === void 0 ? void 0 : options.enabled),
    notifyOnChangeProps: "tracked"
  });
  const set = (0, _react.useCallback)((newData, setOptions) => {
    if (!path) return;
    return firestore.doc(path).set(newData, setOptions || {});
  }, [path, firestore]);
  const update = (0, _react.useCallback)(newData => {
    if (!path) return;
    return firestore.doc(path).update(newData);
  }, [path, firestore]);
  const deleteDocument = (0, _react.useCallback)(() => {
    return deleteDoc(path); // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [path]);
  const setCache = (0, _react.useCallback)(cachedData => {
    if (!path) return;
    const newData = queryClient.setQueryData(path, prevState => {
      return { ...prevState,
        ...cachedData
      };
    });
    updateCollectionCache(queryClient, path, newData.id || "", newData);
  }, // eslint-disable-next-line react-hooks/exhaustive-deps
  [path]);
  return {
    data,
    status,
    error,
    set,
    update,
    deleteDocument,
    refetch,
    setCache,
    unsubscribe: unsubscribeRef.current
  };
};

exports.useDocument = useDocument;
//# sourceMappingURL=useDocument.js.map