import { useCallback, useEffect, useRef } from "react";
import { useQuery, useQueryClient } from "react-query";
import { collectionCache } from "./Cache";
import { empty } from "./types";
import { useHelpers } from "./useHelpers";
import { useFirestore } from "./Provider";
import { parseDates } from "./utils";

function updateCollectionCache(queryClient, path, docId, data) {
  let collection = path.split(`/${docId}`).filter(Boolean);
  collection = collection.join("/");

  if (collection) {
    collectionCache.getKeysFromCollectionPath(collection).forEach(key => {
      queryClient.setQueryData(key, (currentState = empty.array) => {
        // don't mutate the current state if it doesn't include this doc
        if (!currentState.some(currDoc => currDoc.id && currDoc.id === data.id)) {
          return currentState;
        }

        return currentState.map(document => {
          if (document.id === data.id) {
            return data;
          }

          return document;
        });
      });
    });
  }
}

const createListenerAsync = async (firestore, queryClient, path) => {
  if (!path) {
    return {
      unsubscribe: empty.function,
      initialData: empty.object
    };
  }

  return await new Promise((resolve, reject) => {
    const unsubscribe = firestore.doc(path).onSnapshot({
      includeMetadataChanges: true
    }, doc => {
      var _doc$data;

      const docData = (_doc$data = doc.data()) !== null && _doc$data !== void 0 ? _doc$data : empty.object;
      parseDates(docData);
      const data = { ...docData,
        id: doc.id,
        exists: doc.exists,
        hasPendingWrites: doc.metadata.hasPendingWrites
      };

      if (!data.hasPendingWrites) {
        queryClient.setQueryData(path, data);
        updateCollectionCache(queryClient, path, doc.id, data);
        resolve({
          initialData: data,
          unsubscribe
        });
      }
    }, reject);
  });
};

export const useDocument = (path, options) => {
  const {
    firestore
  } = useFirestore();
  const {
    deleteDocument: deleteDoc
  } = useHelpers();
  const queryClient = useQueryClient();
  const unsubscribeRef = useRef(null);
  useEffect(() => {
    return () => {
      // clean up listener on unmount if it exists
      if (unsubscribeRef.current) {
        unsubscribeRef.current();
        unsubscribeRef.current = null;
      }
    };
  }, [path]);

  async function fetch() {
    if (unsubscribeRef.current) {
      unsubscribeRef.current();
      unsubscribeRef.current = null;
    }

    const listner = await createListenerAsync(firestore, queryClient, path);
    unsubscribeRef.current = listner.unsubscribe;
    return listner.initialData;
  }

  const {
    data,
    status,
    error,
    refetch
  } = useQuery(path || "", fetch, { ...options,
    enabled: !!path && (options === null || options === void 0 ? void 0 : options.enabled),
    notifyOnChangeProps: "tracked"
  });
  const set = useCallback((newData, setOptions) => {
    if (!path) return;
    return firestore.doc(path).set(newData, setOptions || {});
  }, [path, firestore]);
  const update = useCallback(newData => {
    if (!path) return;
    return firestore.doc(path).update(newData);
  }, [path, firestore]);
  const deleteDocument = useCallback(() => {
    return deleteDoc(path); // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [path]);
  const setCache = useCallback(cachedData => {
    if (!path) return;
    const newData = queryClient.setQueryData(path, prevState => {
      return { ...prevState,
        ...cachedData
      };
    });
    updateCollectionCache(queryClient, path, newData.id || "", newData);
  }, // eslint-disable-next-line react-hooks/exhaustive-deps
  [path]);
  return {
    data,
    status,
    error,
    set,
    update,
    deleteDocument,
    refetch,
    setCache,
    unsubscribe: unsubscribeRef.current
  };
};
//# sourceMappingURL=useDocument.js.map